defmodule ClickHouse.Interface.HTTP do
  @moduledoc """
  An interface to interact with a ClickHouse server via HTTP.
  """

  @behaviour ClickHouse.Interface

  use GenServer

  require Logger

  alias ClickHouse.{
    Result,
    ResultStream,
    ResultStreamChunk
  }

  alias ClickHouse.Interface.HTTP.Client

  @opts_schema %{
    name: [type: :atom, required: true, default: :default],
    urls: [type: {:list, :binary}, required: true, default: ["http://localhost:8123"]],
    ping_retry: [type: :integer, required: true, default: 3_000],
    pool_timeout: [type: :integer, required: true, default: 150_000],
    pool_max_connections: [type: :integer, required: true, default: 50]
  }

  @ping_statement "SELECT 1"

  ################################
  # ClickHouse.Interface Callbacks
  ################################

  @doc """
  Starts the HTTP network interface.
  """
  @impl ClickHouse.Interface
  @spec start_link(keyword()) :: GenServer.on_start()
  def start_link(opts) do
    opts = KeywordValidator.validate!(opts, @opts_schema)
    GenServer.start_link(__MODULE__, opts)
  end

  @doc """
  Executes a query via HTTP.
  """
  @impl ClickHouse.Interface
  @spec execute(ClickHouse.Client.t(), ClickHouse.Query.t(), keyword()) ::
          {:ok, ClickHouse.Result.t()} | {:error, ClickHouse.error()}
  def execute(client, query, opts \\ []) do
    with {:ok, {pool, urls}} <- get_config(client) do
      opts = Keyword.put(opts, :stream, false)

      case Client.request(pool, urls, query.statement, opts) do
        {:ok, {body, format, meta, compressed}} ->
          result = Result.new(client, body, format, meta, compressed)
          {:ok, result}

        {:error, _} = error ->
          error
      end
    end
  end

  @doc """
  Starts a query result stream via HTTP.
  """
  @impl ClickHouse.Interface
  @spec stream_start(ClickHouse.Client.t(), ClickHouse.Query.t(), keyword()) ::
          {:ok, ClickHouse.ResultStream.t()} | {:error, ClickHouse.error()}
  def stream_start(client, query, opts \\ []) do
    with {:ok, {pool, urls}} <- get_config(client) do
      opts = Keyword.put(opts, :stream, true)

      with {:ok, {ref, opts}} <- Client.request(pool, urls, query.statement, opts) do
        stream = %ResultStream{id: ref, opts: opts}
        {:ok, stream}
      end
    end
  end

  @doc """
  Streams the next results of a query via HTTP.
  """
  @impl ClickHouse.Interface
  @spec stream_next(ClickHouse.Client.t(), ClickHouse.ResultStream.t()) ::
          {:cont, ClickHouse.ResultStream.t()}
          | {:cont, ClickHouse.ResultStream.t(), ClickHouse.ResultStreamChunk.t()}
          | {:halt, ClickHouse.ResultStream.t()}
          | {:error, ClickHouse.error()}
  def stream_next(_client, stream) do
    case Client.stream_next(stream.id, stream.opts) do
      :begin ->
        {:cont, stream}

      {:headers, _} ->
        {:cont, stream}

      {:chunk, data} ->
        stream = %{stream | chunks: stream.chunks + 1}
        chunk = %ResultStreamChunk{id: stream.id, data: data}
        {:cont, stream, chunk}

      :halt ->
        {:halt, stream}

      {:error, _} = error ->
        error
    end
  end

  ################################
  # GenServer Callbacks
  ################################

  @impl GenServer
  def init(opts) do
    state = init_state(opts)
    init_config(state)
    init_pool(state)
    {:ok, state, {:continue, :ping}}
  end

  @impl GenServer
  def handle_continue(:ping, state) do
    ping(state)
    {:noreply, state}
  end

  @impl GenServer
  def handle_info(:retry_ping, state) do
    {:noreply, state, {:continue, :ping}}
  end

  ################################
  # Private API
  ################################

  defp init_state(opts) do
    name = Keyword.get(opts, :name)

    %{
      name: name,
      table_name: table_name(name),
      pool_name: pool_name(name),
      pool_opts: [
        timeout: Keyword.get(opts, :pool_timeout),
        max_connections: Keyword.get(opts, :pool_max_connections)
      ],
      urls: Keyword.get(opts, :urls),
      ping_retry: Keyword.get(opts, :ping_retry)
    }
  end

  defp init_config(state) do
    config = {state.pool_name, state.urls}
    :ets.new(state.table_name, [:named_table, read_concurrency: true])
    :ets.insert(state.table_name, {:config, config})
  end

  defp get_config(client) do
    [{:config, config}] = :ets.lookup(table_name(client.name), :config)
    {:ok, config}
  rescue
    ArgumentError ->
      raise ArgumentError, "No client interface available"
  end

  defp init_pool(state) do
    :hackney_pool.start_pool(state.pool_name, state.pool_opts)
  end

  defp table_name(name) do
    :"#{__MODULE__}.#{name}"
  end

  defp pool_name(name) do
    :"#{__MODULE__}.#{name}.http_pool"
  end

  defp ping(state) do
    case Client.request(state.pool_name, state.urls, @ping_statement, []) do
      {:ok, _} ->
        :ok

      {:error, error} ->
        Logger.error("[ClickHouse.Interface.HTTP] Ping error: #{inspect(error)}")
        Process.send_after(self(), :retry_ping, state.ping_retry)
        :error
    end
  end
end
