defmodule ClickHouse.Query do
  @moduledoc """
  A ClickHouse query.
  """

  alias ClickHouse.{DataType, Query.Statement}

  defstruct [
    :format,
    :statement
  ]

  @typedoc """
  A ClickHouse query
  """
  @type t :: %__MODULE__{
          format: ClickHouse.Format.t() | nil,
          statement: iolist()
        }

  ################################
  # Public API
  ################################

  @doc """
  Prepares a new ClickHouse query.
  """
  @spec prepare(ClickHouse.Client.t(), ClickHouse.statement(), ClickHouse.params()) ::
          ClickHouse.Query.t()
  def prepare(client, statement, params \\ [])

  def prepare(client, statement, params) when is_binary(statement) do
    statement = Statement.new(statement)
    prepare(client, statement, params)
  end

  def prepare(client, %Statement{} = statement, params) do
    %__MODULE__{}
    |> put_format(client.formats, statement.format)
    |> put_statement(statement, params)
  end

  ################################
  # Private API
  ################################

  defp put_format(query, _formats, nil) do
    query
  end

  defp put_format(query, formats, format) do
    case Map.fetch(formats, format) do
      {:ok, format} ->
        %{query | format: format}

      _ ->
        raise ArgumentError, """
        Format '#{format}' not implemented. \
        Please see ClickHouse.Format documentation for more information.
        """
    end
  end

  defp put_statement(query, %Statement{command: :insert, tokens: [statement, _]}, [rows]) do
    params = query.format.encode(rows)
    %{query | statement: [statement, params]}
  end

  defp put_statement(query, %Statement{command: :insert} = statement, []) do
    %{query | statement: statement.tokens}
  end

  defp put_statement(_query, %Statement{command: :insert}, [_, _ | _]) do
    raise ArgumentError, """
    Insert query must contain a single parameter representing all rows.
    """
  end

  defp put_statement(query, %Statement{command: command} = statement, params)
       when command in [:select, :create, :alter, :unknown] do
    statement = encode(statement.tokens, params)
    %{query | statement: statement}
  end

  defp encode(parts, params, acc \\ [])

  defp encode([], _, acc), do: acc

  defp encode(["?" | _], [], _acc) do
    raise ArgumentError, """
    Query parameters do not match the number of embedded '?'.
    """
  end

  defp encode(["?"], [data], acc) do
    [acc | [DataType.encode(data)]]
  end

  defp encode(["?" | parts], [data | params], acc) do
    encode(parts, params, [acc | [DataType.encode(data)]])
  end

  defp encode([part], _params, acc) do
    [acc | [part]]
  end

  defp encode([part | parts], params, acc) do
    encode(parts, params, [acc | [part]])
  end

  defimpl String.Chars do
    def to_string(%{statement: statement}) do
      IO.iodata_to_binary(statement)
    end
  end

  defimpl Inspect do
    def inspect(%{statement: statement}, _) do
      "#ClickHouse.Query<\"#{IO.iodata_to_binary(statement)}\">"
    end
  end
end
